%{
{*******************************************************************************
                                ExpYacc
[CZX]
  {\tgEFA LGPL2(not 2.1) ŔzzĂ܂B

  Eۏ
    ύX҂͖{\tgEFA̎gpɋN邠鑹QEQɂĂ̐ӔC
    ؕȂ̂Ƃ܂B܂oOC̋`𕉂Ȃ̂Ƃ܂B
  Egp
    pA񏤗p킸RɎgpĂ܂܂Bgp\Kv܂B
  EA[JCu̕ύX
    ƎA[JCuɊ܂߂čĔzzĂ܂܂B𓾂Kv܂B
  Eϔłzz
    ȉ[ύX]ɕύX_AύX҂L̏ LGPL2 ŔzzĂB

  EĂ͂Ȃ
    {\tgEFA܂͉ϔłCZXύX(LGPL2 ȊO)zz邱ƁB

[ύX]
v1.00   2002/09/29 t_kumagai@ot.olympus.co.jp(http://kuma.webj.net/)
  EVK쐬
*******************************************************************************}

unit ExpYacc;

interface

uses SysUtils, LexLib, YaccLib, Classes, Math, ExpUtils;

const
  E_UNDEFINED_IDENT = 'Undefined Ident:';	// `̕ϐ
  E_INVALID_FUNCTION = 'Invalid function:';	// Ȋ֐Ăяo
  E_INVALID_PARAM = 'Invalid parameter:';	// ֐p[^Ⴄ
  E_ZERO_DIVIDE = 'Zero Divide:';			// [ɂ鏜Z
  E_VALUE_NOT_ENOUGH = 'Value not enouth:';	// ls
  E_CIRCULATION = 'Circulation:';			// zQ


%}

%token <Double> NUMBER
%token <PChar> IDENT
%type <Double> expr
%type <Integer> args
%type <Integer> arg.list

%left	'+' '-'
%left	'*' '/'
%right	'^'
%right	UNARYOP

%token ILLEGAL 		/* illegal token */

%%

%{
  private
    ValueStack: array of Double;
    FunctionList: TStrings;
    UserList: TStrings;
    ValueList: TStrings;
    CycleList: TStrings;
  protected
    // G[ override
    procedure yyerror(Msg: string); override;
    procedure YYAssert(Cond: Boolean; Msg: string);
    function SetValue(Value: Double): Double;
    function IdentToValue(const ID: string): Double;
    procedure PushStack(Value: Double);
    function PopStack: Double;
    function TopStack: PDouble;
    function CountStack: Integer;
    function OpUserFunc(FuncName, FuncDef: string; N: Integer; P: PDouble): Double;
    function CallFunction(const ID: string; ArgCount: Integer): Double;
  public
    Value: Double;
    procedure Init(VL, UL, CL: TStrings; MaxStackSize: Integer = 64);
    procedure Term;
%}

start	: expr 							{ SetValue($1); }
		| error							{ }
		;

expr	: expr '+' expr					{ $$ := $1 + $3; }
		| expr '-' expr					{ $$ := $1 - $3; }
		| expr '*' expr					{ $$ := $1 * $3; }
		| expr '/' expr					{
											YYAssert($3 <> 0, E_ZERO_DIVIDE + yylex.yytext);
											$$ := $1 / $3;
										}
		| expr '^' expr					{ $$ := Power($1, $3); }
		| '-' expr	%prec UNARYOP		{ $$ := -$2; }
		| '+' expr	%prec UNARYOP		{ $$ := $2; }
		| '(' expr ')'					{ $$ := $2; }
		| NUMBER						{ $$ := $1; }
		| IDENT							{ $$ := IdentToValue($1); }
		| IDENT '(' args ')'			{ $$ := CallFunction($1, $3); }
		;
args	:								{ $$ := 0; }
		| arg.list						{ $$ := $1; }
		;

arg.list
		: expr							{ $$ := 1; PushStack($1); }
		| expr ',' arg.list				{ Inc($$); PushStack($1); }
		;
%%

// G[ override
procedure TYacc.yyerror (Msg : String );
var
  S: string;
begin
  S := yylex.yytext;
  yylex.yywriteln(Msg +
    ' around ''' + S + '''' +
    ' at L' + IntToStr(yylex.yylineno) +
    ':C' + IntToStr(yylex.yycolno));
  inherited yyerror(Msg);
end;

procedure YYAssert(Cond: Boolean; Msg: string);
begin
  if (Cond) then
    Exit;
  raise Exception.Create(Msg);
end;

procedure TYacc.YYAssert(Cond: Boolean; Msg: string);
begin
  ExpYacc.YYAssert(Cond, Msg);
end;

function TYacc.SetValue(Value: Double): Double;
begin
  Self.Value := Value;
  Result := Value;
end;

function TYacc.IdentToValue(const ID: string): Double;
begin
  YYAssert(ValueList.Values[ID] <> '', E_UNDEFINED_IDENT + ID);
  YYAssert(CycleList.IndexOf(ID) = -1, E_CIRCULATION);
  CycleList.Insert(0, ID);
  Result := CustomExpToFloat(ValueList.Values[ID], ValueList, UserList, CycleList);
  CycleList.Delete(0);
end;

procedure TYacc.PushStack(Value: Double);
var
  N: Integer;
begin
  N := CountStack;
  SetLength(ValueStack, N + 1);
  ValueStack[N] := Value;
end;

function TYacc.PopStack: Double;
var
  N: Integer;
begin
  N := CountStack;
  Result := ValueStack[N - 1];
  SetLength(ValueStack, N - 1);
end;

function TYacc.TopStack: PDouble;
begin
  Result := @ValueStack[CountStack - 1];
end;

function TYacc.CountStack: Integer;
begin
  Result := Length(ValueStack);
end;

function OpSum(N: Integer; P: PDouble): Double;
var
  I: Integer;
begin
  Result := 0;
  for I := 0 to N - 1 do
  begin
    Result := Result + P^;
    Inc(P);
  end;
end;

function OpAverage(N: Integer; P: PDouble): Double;
begin
  YYAssert(N > 0, E_VALUE_NOT_ENOUGH);
  Result := OpSum(N, P) / N;
end;

function OpStdev(N: Integer; P: PDouble): Double;
var
  S, S2: Double;
  I: Integer;
begin
  YYAssert(N > 1,  E_VALUE_NOT_ENOUGH);
  S := 0;
  S2 := 0;
  for I := 0 to N - 1 do
  begin
    S := S + P^;
    S2 := S2 + Sqr(P^);
    Inc(P);
  end;
  Result := Sqrt(N * S2 - Sqr(S)) / (N * (N - 1));
end;

function OpAbs(N: Integer; P: PDouble): Double;
begin
  YYAssert(N = 1, E_INVALID_PARAM);
  Result := Abs(P^);
end;

function OpSin(N: Integer; P: PDouble): Double;
begin
  YYAssert(N = 1, E_INVALID_PARAM);
  Result := Sin(P^);
end;

function OpCos(N: Integer; P: PDouble): Double;
begin
  YYAssert(N = 1, E_INVALID_PARAM);
  Result := Cos(P^);
end;

function OpTan(N: Integer; P: PDouble): Double;
begin
  YYAssert(N = 1, E_INVALID_PARAM);
  Result := Tan(P^);
end;

function OpASin(N: Integer; P: PDouble): Double;
begin
  YYAssert(N = 1, E_INVALID_PARAM);
  YYAssert((-1 <= P^) and (P^ <= 1), E_INVALID_PARAM);
  Result := ArcSin(P^);
end;

function OpACos(N: Integer; P: PDouble): Double;
begin
  YYAssert(N = 1, E_INVALID_PARAM);
  YYAssert((-1 <= P^) and (P^ <= 1), E_INVALID_PARAM);
  Result := ArcCos(P^);
end;

function OpATan(N: Integer; P: PDouble): Double;
begin
  YYAssert(N = 1, E_INVALID_PARAM);
  Result := ArcTan(P^);
end;

function OpATan2(N: Integer; P: PDouble): Double;
var
  X, Y: Double;
begin
  YYAssert(N = 2, E_INVALID_PARAM);
  Y := P^;
  Inc(P);
  X := P^;
  Result := ArcTan2(Y, X);
end;

function OpExp(N: Integer; P: PDouble): Double;
begin
  YYAssert(N = 1, E_INVALID_PARAM);
  Result := Exp(P^);
end;

function OpMax(N: Integer; P: PDouble): Double;
var
  I: Integer;
begin
  YYAssert(N > 0, E_VALUE_NOT_ENOUGH);
  Result := P^;
  Inc(P);
  for I := 1 to N - 1 do
  begin
    Result := Max(Result, P^);
    Inc(P);
  end;
end;

function OpMin(N: Integer; P: PDouble): Double;
var
  I: Integer;
begin
  YYAssert(N > 0, E_VALUE_NOT_ENOUGH);
  Result := P^;
  Inc(P);
  for I := 1 to N - 1 do
  begin
    Result := Min(Result, P^);
    Inc(P);
  end;
end;

function OpSqrt(N: Integer; P: PDouble): Double;
begin
  YYAssert(N = 1, E_INVALID_PARAM);
  YYAssert(P^ >= 0, E_INVALID_PARAM);
  Result := Sqrt(P^);
end;

function OpLog(N: Integer; P: PDouble): Double;
begin
  YYAssert(N = 1, E_INVALID_PARAM);
  Result := Ln(P^);
end;

function OpLog10(N: Integer; P: PDouble): Double;
begin
  YYAssert(N = 1, E_INVALID_PARAM);
  Result := Log10(P^);
end;

function OpSinh(N: Integer; P: PDouble): Double;
begin
  YYAssert(N = 1, E_INVALID_PARAM);
  Result := Sinh(P^);
end;

function OpCosh(N: Integer; P: PDouble): Double;
begin
  YYAssert(N = 1, E_INVALID_PARAM);
  Result := Cosh(P^);
end;

function OpTanh(N: Integer; P: PDouble): Double;
begin
  YYAssert(N = 1, E_INVALID_PARAM);
  Result := Tanh(P^);
end;

function TYacc.OpUserFunc(FuncName, FuncDef: string; N: Integer; P: PDouble): Double;
// FuncDef: 'A, B','A+B'
var
  FD, Args: TStrings;
  I: Integer;
begin
  YYAssert(CycleList.IndexOf(FuncName) = -1, E_CIRCULATION);
  FD := TStringList.Create;
  Args := TStringList.Create;
  try
    FD.CommaText := FuncDef;
    with (TStringList.Create) do
    try
      CommaText := FD[0];
      YYAssert(Count = N, E_INVALID_PARAM);
      Args.AddStrings(ValueList);
      for I := N - 1 downto 0 do
      begin
        Args.Values[Strings[I]] := FloatToStr(P^);
        Inc(P);
      end;
      CycleList.Insert(0, FuncName);
      Result := CustomExpToFloat(FD[1], Args, UserList, CycleList);
      CycleList.Delete(0);
    finally
      Free;
    end;
  finally
    FD.Free;
    Args.Free;
  end;
end;

procedure TYacc.Init(VL, UL, CL: TStrings; MaxStackSize: Integer);
begin
  Assert(Assigned(VL));
  Assert(Assigned(UL));
  Assert(Assigned(CL));
  ValueList := VL;
  UserList := UL;
  CycleList := CL;
  SetLength(ValueStack, 0);

  FunctionList := TStringList.Create;
  with (FunctionList) do
  begin
    AddObject('SUM', @OpSum);
    AddObject('AVERAGE', @OpAverage);
    AddObject('STDEV', @OpStdev);
    AddObject('ABS', @OpAbs);
    AddObject('SIN', @OpSin);
    AddObject('COS', @OpCos);
    AddObject('TAN', @OpTan);
    AddObject('ASIN', @OpASin);
    AddObject('ACOS', @OpACos);
    AddObject('ATAN', @OpATan);
    AddObject('ATAN2', @OpATan2);
    AddObject('EXP', @OpExp);
    AddObject('MAX', @OpMax);
    AddObject('MIN', @OpMin);
    AddObject('SQRT', @OpSqrt);
    AddObject('LOG', @OpLog);
    AddObject('LOG10', @OpLog10);
    AddObject('SINH', @OpSinh);
    AddObject('COSH', @OpCosh);
    AddObject('TANH', @OpTanh);
  end;
end;

procedure TYacc.Term;
begin
  ValueList := nil;
  UserList := nil;
  CycleList := nil;
  SetLength(ValueStack, 0);
  FunctionList.Free;
end;

function TYacc.CallFunction(const ID: string; ArgCount: Integer): Double;
type
  TFuncCall = function (N: Integer; P: PDouble): Double;
var
  Index: Integer;
begin
  if (UserList.Values[ID] <> '') then
  begin
    Result := OpUserFunc(ID + '()', UserList.Values[ID], ArgCount, @ValueStack[CountStack - ArgCount]);
  end else
  begin
    Index := FunctionList.IndexOf(ID);
    YYAssert(Index >= 0, E_INVALID_FUNCTION + ID);
    Result := TFuncCall(FunctionList.Objects[Index])(ArgCount, @ValueStack[CountStack - ArgCount]);
  end;
  SetLength(ValueStack, CountStack - ArgCount);
end;

end.
